%************************************************************************
% LSSVM_DAE_Demo: 
%
% Learning solution of Descriptor Systems using Least Squares Support Vector Machines
%
% Created by
%     Siamak Mehrkanoon
%     Dept. of Electrical Engineering (ESAT)
%     Research Group: STADIUS
%     KU LEUVEN
%
% (c) 2012
%************************************************************************

% Citations:

%[1] Mehrkanoon S., Falck T., Suykens J.A.K., 
%"Approximate Solutions to Ordinary Differential Equations Using Least Squares Support Vector Machines",
%IEEE Transactions on Neural Networks and Learning Systems, vol. 23, no. 9, Sep. 2012, pp. 1356-1367.


%[2] Mehrkanoon S., Suykens J.A.K.,
%"LS-SVM approximate solution to linear time varying descriptor systems", 
%Automatica, vol. 48, no. 10, Oct. 2012, pp. 2502-2511.


%[3] Mehrkanoon S., Suykens J.A.K., 
%"Learning Solutions to Partial Differential Equations using LS-SVM",
%Neurocomputing, vol. 159, Mar. 2015, pp. 105-116.


%Author: Siamak Mehrkanoon

%%  ====== An Index 3 DAE, Problem 5.2 in reference [2]  ==========

%  C dX/dt = A X + B U (index-3 DAE)
%  
%  X=[x_1 x_2 X_3]
%  U=[u_1 u_2 u_3]
%  C= [           A=[            X=[
%      0 -t 0         -1 0 0        x_1
%      1 0 t           0 -1 0       x_2 
%      0 1 0]          0 0 -1]      x_3]
%
% B=[0 0 0 , 0 0 0, 0 0 0];
% 
%
% 0 <=  t  < = 20
%
% Initial Condition
% X(0) = [0, e^-1, e^-1];


%% ===================================================
clear all; close all; clc
warning('off','all');



% ============= Exact solution =============
ex1=@(x)  -x.*exp(-(x+1));
ex2=@(x) exp(-(x+1));
ex3=@(x) exp(-(x+1));


% ==========  time interval =============
t0=0;
tf=21;

%========== initial conditions =============
a1=0;  
a2=1/exp(1);
a3=1/exp(1);

% =========================================
np=70;  % Number of training points
x= linspace(t0,tf,np);
s=size(x,2);
h= (tf-t0)/np;
pt_test=linspace(t0,tf-1,200);
pt_test=pt_test';  % test points

% ============== model parameters ==============

gamma=10^12;
ng_gam=50;
sigma_range = logspace(-2,3,ng_gam);


% ==================== Matrix A =================

a11= -1*ones(s-1,1);
a12= 0*ones(s-1,1);
a13= 0*ones(s-1,1);

a21= 0*ones(s-1,1);
a22= -1*ones(s-1,1);
a23= 0*ones(s-1,1);

a31= 0*ones(s-1,1);
a32=0*ones(s-1,1);
a33= -1*ones(s-1,1);

% ======================== Matrix C ===============

c11= 0*ones(s-1,1);
c12=@(x) -x;
c13=0*ones(s-1,1);

c21= ones(s-1,1);
c22= 0*ones(s-1,1);
c23=@(x) x;

c31= 0*ones(s-1,1);
c32 = ones(s-1,1);
c33= 0*ones(s-1,1);


% ================  [g1 g2 g3 ]= B U  ===========================

g1= 0*ones(s-1,1);   
g2=0*ones(s-1,1);
g3= 0*ones(s-1,1);
x=x';
pt_training=x(2:end);   % Training points
pt_validation= pt_training(2:end)-(h/2);  % validation points
m=3;   % number of equations


%%  
% =============== DA =================

DA11=diag(a11);
DA12=diag(a12);
DA13=diag(a13);


DA21=diag(a21);
DA22=diag(a22);
DA23=diag(a23);



DA31=diag(a31);
DA32=diag(a32);
DA33=diag(a33);


% ============= DC ====================

DC11=diag(c11);
DC12=diag(c12(pt_training));
DC13=diag(c13);

DC21=diag(c21);
DC22=diag(c22);
DC23=diag(c23(pt_training));

DC31=diag(c31);
DC32=diag(c32);
DC33=diag(c33);


% =============  F ====================

F11=a11;
F12=a12;
F13=a13;

F21=a21;
F22=a22;
F23=a23;

F31=a31;
F32=a32;
F33=a33;


% ====================================
BB=zeros(ng_gam,1);
for sig_idx=1:size(sigma_range,2)
    
    sig = sigma_range(sig_idx);
    K=KernelMatrix(pt_training,'RBF_kernel',sig);
    m1=size(K,1);
    xx1=pt_training*ones(1,size(pt_training,1));
    xx2=pt_training*ones(1,size(pt_training,1));
    cof1=-2*(xx1-xx2')/(sig);
    Kx=(cof1.*K)';
    Ky=(-cof1.*K)';
    Kxy=(2/sig)*K - (cof1.^2) .* K;
    
    omega00=[];
    omega11=[];
    omega10=[];
    omega01=[];
    
    for i=1:m
        omega00=blkdiag(omega00,K);
        omega11=blkdiag(omega11,Kxy);
        omega10=blkdiag(omega10,Kx);
        omega01=blkdiag(omega01,Ky);
    end  
    DCbar1=[DC11 DC12 DC13 ];
    DCbar2=[DC21 DC22 DC23 ];
    DCbar3=[DC31 DC32 DC33 ];
    
    DAbar1=[DA11 DA12 DA13];
    DAbar2=[DA21 DA22 DA23 ];
    DAbar3=[DA31 DA32 DA33 ];

    K11=DCbar1*omega11*DCbar1' - DAbar1*omega10*DCbar1' - DCbar1 * omega01 * DAbar1' + DAbar1 * omega00 * DAbar1';
    K12=DCbar1*omega11*DCbar2' - DAbar1*omega10*DCbar2' - DCbar1 * omega01 * DAbar2' + DAbar1 * omega00 * DAbar2';
    K13=DCbar1*omega11*DCbar3' - DAbar1*omega10*DCbar3' - DCbar1 * omega01 * DAbar3' + DAbar1 * omega00 * DAbar3';
      
    K21=DCbar2*omega11*DCbar1' - DAbar2*omega10*DCbar1' - DCbar2 * omega01 * DAbar1' + DAbar2 * omega00 * DAbar1';
    K22=DCbar2*omega11*DCbar2' - DAbar2*omega10*DCbar2' - DCbar2 * omega01 * DAbar2' + DAbar2 * omega00 * DAbar2';
    K23=DCbar2*omega11*DCbar3' - DAbar2*omega10*DCbar3' - DCbar2 * omega01 * DAbar3' + DAbar2 * omega00 * DAbar3';
  
    K31=DCbar3*omega11*DCbar1' - DAbar3*omega10*DCbar1' - DCbar3 * omega01 * DAbar1' + DAbar3 * omega00 * DAbar1';
    K32=DCbar3*omega11*DCbar2' - DAbar3*omega10*DCbar2' - DCbar3 * omega01 * DAbar2' + DAbar3 * omega00 * DAbar2';
    K33=DCbar3*omega11*DCbar3' - DAbar3*omega10*DCbar3' - DCbar3 * omega01 * DAbar3' + DAbar3 * omega00 * DAbar3';
 
    Omega= [K11 K12 K13 ;...
        K21 K22 K23 ;...
        K31 K32 K33 ];
    m2=size(Omega,1);
    K2=KernelMatrix(pt_training,'RBF_kernel',sig,x(1));
    xx1=pt_training*ones(1,size(x(1),1));
    xx2=x(1)*ones(1,size(pt_training,1));
    cof2=-2*(xx1-xx2')/(sig);
    
    S11= DC11*cof2.*K2-DA11*K2;
    S12= DC12*cof2.*K2-DA12*K2;
    S13= DC13*cof2.*K2-DA13*K2;
    
    S21= DC21*cof2.*K2-DA21*K2;
    S22= DC22*cof2.*K2-DA22*K2;
    S23= DC23*cof2.*K2-DA23*K2;
    
    S31= DC31*cof2.*K2-DA31*K2;
    S32= DC32*cof2.*K2-DA32*K2;
    S33= DC33*cof2.*K2-DA33*K2;
    
    S=[S11, S12, S13 ;...
        S21,S22, S23 ;...
        S31, S32, S33];
    
    F=[-F11 , -F12, -F13;...
        -F21 , -F22, -F23;...
        -F31, -F32, -F33];
    
    A=[ Omega+1/gamma*eye(m2) ,  S , F ;...
        S' , eye(m), eye(m) ;...
        F', eye(m) , zeros(m)];
     B=[g1; g2; g3; a1 ; a2 ; a3;   0 ; 0 ;0 ];
    result=A\B;
    alpha_1=result(1:m1);
    alpha_2=result(m1+1:2*m1);
    alpha_3=result(2*m1+1:3*m1);
    beta_1=result(3*m1+1);
    beta_2=result(3*m1+2);
    beta_3=result(3*m1+3);
    b_1=result(end-2);
    b_2=result(end-1);
    b_3=result(end);

    % =================================
    
    Kt=KernelMatrix(pt_training,'RBF_kernel',sig,pt_validation);
    xx1=pt_training*ones(1,size(pt_validation,1));
    xx2=pt_validation*ones(1,size(pt_training,1));
    cof3=-2*(xx1-xx2')/(sig);
    Kxt= cof3 .* Kt;
    K2t=KernelMatrix(pt_validation,'RBF_kernel',sig,x(1));
    yUval_1=(Kxt'*DC11 - Kt'*DA11 )* alpha_1 + (Kxt'*DC21-Kt'*DA21)*alpha_2 + (Kxt'*DC31-Kt'*DA31)*alpha_3   +  K2t* beta_1  + b_1;
    yUval_2=(Kxt'*DC22 - Kt'*DA22 )* alpha_2 + (Kxt'*DC12-Kt'*DA12)*alpha_1  + (Kxt'*DC32-Kt'*DA32)*alpha_3   +  K2t* beta_2  + b_2;
    yUval_3=(Kxt'*DC33 - Kt'*DA33 )* alpha_3 + (Kxt'*DC13-Kt'*DA13)*alpha_1  + (Kxt'*DC23-Kt'*DA23)*alpha_2  +  K2t* beta_3  + b_3;  
    xx3=pt_validation*ones(1,size(x(1),1));
    xx4=x(1)*ones(1,size(pt_validation,1));
    cof4=2*(xx3-xx4')/(sig);
    K1yt = - cof4 .* K2t;
    Kxyt=(  ( (2/sig) - (cof3.^2) ) .* Kt);
    Kyt= - cof3 .* Kt ;
    yprimeUval_1=(Kxyt'*DC11  - Kyt'*DA11 )* alpha_1  + (Kxyt'*DC21 - Kyt'*DA21)*alpha_2 +  (Kxyt'*DC31 - Kyt'*DA31)*alpha_3   +    K1yt* beta_1 ;
    yprimeUval_2=(Kxyt'*DC22  - Kyt'*DA22 )* alpha_2  + (Kxyt'*DC12- Kyt'*DA12)*alpha_1   + (Kxyt'*DC32- Kyt'*DA32)*alpha_3 +   + K1yt* beta_2 ;
    yprimeUval_3=(Kxyt'*DC33  - Kyt'*DA33 )* alpha_3  + (Kxyt'*DC13- Kyt'*DA13)*alpha_1   + (Kxyt'*DC23- Kyt'*DA23)*alpha_2  +  K1yt* beta_3 ;
    e_1=  diag(c12(pt_validation))*yprimeUval_2 + yUval_1;
    e_2=  yprimeUval_1  + diag(c23(pt_validation))*yprimeUval_3  + yUval_2;
    e_3= yprimeUval_2  + yUval_3 ;
    E=[e_1 , e_2, e_3];
    BB(sig_idx)=max(max(abs(E)));
    
end
[minBB idx] = min(BB);
sig=sigma_range(idx);
fprintf('------------------------------------------\n\n')
fprintf('Optimal sigma (kernel bandwidth)=%f\n\n',sig)

%%  ============== Retrain =======================

K=KernelMatrix(pt_training,'RBF_kernel',sig);
m1=size(K,1);
xx1=pt_training*ones(1,size(pt_training,1));
xx2=pt_training*ones(1,size(pt_training,1));
cof1=-2*(xx1-xx2')/(sig);
Kx=(cof1.*K)';
Ky=(-cof1.*K)';
Kxy=(2/sig)*K - (cof1.^2) .* K;
omega00=[];
omega11=[];
omega10=[];
omega01=[];
for i=1:m
    omega00=blkdiag(omega00,K);
    omega11=blkdiag(omega11,Kxy);
    omega10=blkdiag(omega10,Kx);
    omega01=blkdiag(omega01,Ky);
end
DCbar1=[DC11 DC12 DC13 ];
DCbar2=[DC21 DC22 DC23 ];
DCbar3=[DC31 DC32 DC33 ];

DAbar1=[DA11 DA12 DA13];
DAbar2=[DA21 DA22 DA23 ];
DAbar3=[DA31 DA32 DA33 ];

K11=DCbar1*omega11*DCbar1' - DAbar1*omega10*DCbar1' - DCbar1 * omega01 * DAbar1' + DAbar1 * omega00 * DAbar1';
K12=DCbar1*omega11*DCbar2' - DAbar1*omega10*DCbar2' - DCbar1 * omega01 * DAbar2' + DAbar1 * omega00 * DAbar2';
K13=DCbar1*omega11*DCbar3' - DAbar1*omega10*DCbar3' - DCbar1 * omega01 * DAbar3' + DAbar1 * omega00 * DAbar3';

K21=DCbar2*omega11*DCbar1' - DAbar2*omega10*DCbar1' - DCbar2 * omega01 * DAbar1' + DAbar2 * omega00 * DAbar1';
K22=DCbar2*omega11*DCbar2' - DAbar2*omega10*DCbar2' - DCbar2 * omega01 * DAbar2' + DAbar2 * omega00 * DAbar2';
K23=DCbar2*omega11*DCbar3' - DAbar2*omega10*DCbar3' - DCbar2 * omega01 * DAbar3' + DAbar2 * omega00 * DAbar3';

K31=DCbar3*omega11*DCbar1' - DAbar3*omega10*DCbar1' - DCbar3 * omega01 * DAbar1' + DAbar3 * omega00 * DAbar1';
K32=DCbar3*omega11*DCbar2' - DAbar3*omega10*DCbar2' - DCbar3 * omega01 * DAbar2' + DAbar3 * omega00 * DAbar2';
K33=DCbar3*omega11*DCbar3' - DAbar3*omega10*DCbar3' - DCbar3 * omega01 * DAbar3' + DAbar3 * omega00 * DAbar3';

Omega= [K11 K12 K13 ;...
    K21 K22 K23 ;...
    K31 K32 K33 ];

m2=size(Omega,1);
K2=KernelMatrix(pt_training,'RBF_kernel',sig,x(1));
xx1=pt_training*ones(1,size(x(1),1));
xx2=x(1)*ones(1,size(pt_training,1));
cof2=-2*(xx1-xx2')/(sig);

S11= DC11*cof2.*K2-DA11*K2;
S12= DC12*cof2.*K2-DA12*K2;
S13= DC13*cof2.*K2-DA13*K2;

S21= DC21*cof2.*K2-DA21*K2;
S22= DC22*cof2.*K2-DA22*K2;
S23= DC23*cof2.*K2-DA23*K2;

S31= DC31*cof2.*K2-DA31*K2;
S32= DC32*cof2.*K2-DA32*K2;
S33= DC33*cof2.*K2-DA33*K2;

S=[S11, S12, S13 ;...
    S21,S22, S23 ;...
    S31, S32, S33];

F=[-F11 , -F12, -F13;...
    -F21 , -F22, -F23;...
    -F31, -F32, -F33];

A=[ Omega+1/gamma*eye(m2) ,  S , F ;...
    S' , eye(m), eye(m) ;...
    F', eye(m) , zeros(m)];

B=[g1; g2 ; g3 ; a1 ; a2 ; a3;   0 ; 0 ;0 ];
result=A\B;
alpha_1=result(1:m1);
alpha_2=result(m1+1:2*m1);
alpha_3=result(2*m1+1:3*m1);
beta_1=result(3*m1+1);
beta_2=result(3*m1+2);
beta_3=result(3*m1+3);
b_1=result(end-2);
b_2=result(end-1);
b_3=result(end);


%% Test set 
Kt=KernelMatrix(pt_training,'RBF_kernel',sig,pt_test);
xx1=pt_training*ones(1,size(pt_test,1));
xx2=pt_test*ones(1,size(pt_training,1));
cof3=-2*(xx1-xx2')/(sig);
Kxt= cof3 .* Kt;
K2t=KernelMatrix(pt_test,'RBF_kernel',sig,x(1));
ytest_1=(Kxt'*DC11 - Kt'*DA11 )* alpha_1 + (Kxt'*DC21 - Kt'*DA21)*alpha_2 + (Kxt'*DC31 - Kt'*DA31)*alpha_3   +  K2t* beta_1  + b_1;
ytest_2=(Kxt'*DC22 - Kt'*DA22 )* alpha_2 + (Kxt'*DC12 - Kt'*DA12)*alpha_1 + (Kxt'*DC32 - Kt'*DA32)*alpha_3   +  K2t* beta_2  + b_2;
ytest_3=(Kxt'*DC33 - Kt'*DA33 )* alpha_3 + (Kxt'*DC13 - Kt'*DA13)*alpha_1 + (Kxt'*DC23 - Kt'*DA23)*alpha_2   +  K2t* beta_3  + b_3;


ee1=ex1(pt_test)-ytest_1;
ee2=ex2(pt_test)-ytest_2;
ee3=ex3(pt_test)-ytest_3;
MSE1=mse(ee1);
MSE2=mse(ee2);
MSE3=mse(ee3);
fprintf('MSE on test set for x1=%d\n',MSE1)
fprintf('MSE on test set for x2=%d\n',MSE2)
fprintf('MSE on test set for x3=%d\n\n',MSE3)
fprintf('-------  Finished -----------------------\n\n')


% ============= plotting the solutions ====================
size=12;
figure
subplot(1,3,1)
ezplot('-t*exp(-(t+1))',[t0,tf])
hold all
plot(pt_test,ytest_1,'r--')
legend('Exact solution','Approximate solution')
set(gca,'Fontsize',size)
grid on

subplot(1,3,2)
ezplot('exp(-(x+1))',[t0,tf])
hold all
plot(pt_test,ytest_2,'r--')
legend('Exact solution','Approximate solution')
set(gca,'Fontsize',size)
grid on

subplot(1,3,3)
ezplot('exp(-(x+1))',[t0,tf])
hold all
plot(pt_test,ytest_3,'r--')
legend('Exact solution','Approximate solution')
set(gca,'Fontsize',size)
grid on

% =================  plotting the residuals ======================

size=15;
figure
subplot
subplot(1,3,1)
plot(pt_test,ee1,'r')
xlabel('t','Fontsize',size)
ylabel('x1-xhat1','Fontsize',size)
set(gca,'Fontsize',size)
title(['MSE=',num2str(MSE1)],'Fontsize',15)
grid on


subplot(1,3,2)
plot(pt_test,ee2,'r')
xlabel('t','Fontsize',size)
ylabel('x2-xhat2','Fontsize',size)
set(gca,'Fontsize',size)
title(['MSE=',num2str(MSE2)],'Fontsize',15)
grid on

subplot(1,3,3)
plot(pt_test,ee2,'r')
xlabel('t','Fontsize',size)
ylabel('x3-xhat3','Fontsize',size)
set(gca,'Fontsize',size)
title(['MSE=',num2str(MSE3)],'Fontsize',15)
grid on